#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <assert.h>
#include <unistd.h>
#include <sys/time.h>
#include <math.h>
#include "state.h"
#include "io.h"
#include "mooremachine.h"

//Private functions
long long get_time();
static void update_state(moore_t* m);
static void read_inputs(moore_t* m);
static void write_outputs(moore_t* m);

//Main Cycle
void run(moore_t* m, cycle_fun begin_run, void* p)
{
	long long time = 0;
  long long time_diff = 0;
  while(true)
  {
  	time = get_time();
    begin_run(p);
    read_inputs(m);
    update_state(m);
    write_outputs(m);
    if (labs(get_time() - time)*1000 > m->cycle_limit)
    	printf("Exceeded cycle time limit!\n");	 
  }
}

moore_t* create_moore_machine(int num_of_inputs, int num_of_outputs, int num_of_states, long long cycle_limit,  init initialize, void* init_p)
{
	moore_t* m = (moore_t*)malloc(sizeof(moore_t));
	assert(m!=NULL);
	m->num_inputs = num_of_inputs;
	m->num_outputs = num_of_outputs;
	m->num_states = num_of_states;
	m->inputs = (input_t**)malloc(num_of_inputs*sizeof(input_t*));
	m->outputs = (output_t**)malloc(num_of_outputs*sizeof(output_t*));
	m->reads = (read_in*)malloc(num_of_inputs*sizeof(read_in));
	m->writes = (write_out*)malloc(num_of_outputs*sizeof(write_out));
	m->states = (state_t**)malloc(num_of_states*sizeof(state_t*));
	m->cycle_limit = cycle_limit;
	m->current_state = NULL;

  //Additional initialization function 
	initialize(m, init_p);
	return m;
}

void destroy_moore_machine(moore_t* m)
{
	int i;
	if(m)
	{
		for(i = 0; i < m->num_inputs; i++)
			free(m->inputs[i]);
		for(i = 0; i < m->num_outputs; i++)
			free(m->outputs[i]);
		for(i = 0; i < m->num_states; i++)
			destroy_state(m->states[i]);
			
		free(m->inputs);
		free(m->outputs);
		free(m->states);
		free(m->writes);
		free(m->reads);
		free(m);
	}
}

bool register_state(moore_t* m, state_id_t id, state_t* state)
{
	if(m && id < m->num_states)
	{
		m->states[id]=state;
		return true;
	}
	return false;
}

bool set_start_state(moore_t* m, state_id_t id)
{
	if(m && id < m->num_states)
	{
		m->current_state = m->states[id];
		return true;
	}
	return false;
}

bool register_input(moore_t* m, input_id_t id, input_t* in, read_in read)
{
	if(m && id < m->num_inputs)
	{
		m->inputs[id]=in;
		m->reads[id] = read;
		return true;
	}
	return false;
}

bool register_output(moore_t* m, output_id_t id, output_t* out, write_out write)
{
	if(m && id < m->num_outputs)
	{
		m->outputs[id]=out;
		m->writes[id] = write;
		return true;
	}
	return false;
}

void update_state(moore_t* m)
{
	m->current_state = m->current_state->next_state(m->inputs, (void**)m->states, NULL);
}

void write_outputs(moore_t* m)
{
	int i;
	m->current_state->update_output(m->outputs,NULL);
	for(i = 0; i < m->num_outputs; i++)
		m->writes[i](m->outputs[i]);
}

void read_inputs(moore_t* m)
{
  int i;
  for(i = 0; i < m->num_inputs; i++)
		m->reads[i](m->inputs[i]);
}

long long get_time() {
    struct timeval te; 
    gettimeofday(&te, NULL); // get current time
    long long milliseconds = te.tv_sec*1000LL + te.tv_usec/1000; // caculate milliseconds
    return milliseconds;
}
