#ifndef MOORE_MACHINE_H
#define MOORE_MACHINE_H
#include <stdbool.h>
#include "state.h"
#include "io.h"

/*
* The module contains the basic functions and structure to 
* realize a Moore Machine.
*/

//Structure of a Moore Machine
struct moore_s
{
	int num_inputs;           //Number of inputs
	int num_states;           //Number of states
	int num_outputs;          //Number of output
	input_t** inputs;         //Input vector
	state_t** states;         //States vector
	output_t** outputs;       //Output vector
	state_t* current_state;   //Reference to the current state
  long long cycle_limit; //Machine cycle time limit 
	
	//IO specific function
	read_in* reads;           //Vector of read input functions
	write_out* writes;        //Vector of write output functions
};
//Type of a Moore Machine
typedef struct moore_s moore_t;

/*
* Type of function to add additional steps 
* to the machine initialization function
* The parameter of the function are:
*   - Moore machine
*   - Free parameter
*/
typedef void (*init)(moore_t* m, void* p);

/*
* Type of function to add additional steps to the main cycle
* Parameters:
*   - Free parameter
*/
typedef void (*cycle_fun)(void *p);

/*
* Function to create a new Moore Machine
* Parameter:
*   - Number of expected inputs
*   - Number of expected outputs
*   - Number of expected states
*   - State cycle limit
*   - Additional initialization function
*   - Free parameter for the additional initialization function
*/
moore_t* create_moore_machine(int num_of_inputs, int num_of_outputs, int num_of_states, long long cycle_limit, init initialize, void* init_p);

//The function set the fist state of the machine
bool set_start_state(moore_t* m, state_id_t id);

//Function to destroy a machine
void destroy_moore_machine(moore_t* m);

/*
* Function to register a new state:
* The function checks if the identifier is not greater than 
* the number of states provided when the machine has been created
* Parameters:
*   - Moore Machine
*   - State identifier
*   - Reference to the state
*/
bool register_state(moore_t* m, state_id_t id, state_t* state);

/*
* Function to register a new input:
* The function checks if the identifier is not greater than 
* the number of inputs provided when the machine has been created
* Parameters:
*   - Moore Machine
*   - Input identifier
*   - Reference to the input
*   - Function to read the input value
*/
bool register_input(moore_t* m, input_id_t id, input_t* in, read_in read);

/*
* Function to register a new output:
* The function checks if the identifier is not greater than 
* the number of outputs provided when the machine has been created
* Parameters:
*   - Moore Machine
*   - Output identifier
*   - Reference to the output
*   - Function to write the output value
*/
bool register_output(moore_t* m, output_id_t id, output_t* out, write_out write);

/*
* Main function to run the machine
* Properties:
*   - Moore Machine
*   - Additional function to run at the beginning of the cycle
*   - Free parameter of the additional cycle function
*/
void run(moore_t* m, cycle_fun run, void* p);

#endif
