#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <wiringPi.h>
#include "state.h"
#include "io.h"
#include "mooremachine.h"
#include "coffee_machine.h"


static char stdin_buffer[256];
static state_t states[NUM_OF_STATES];
static input_t inputs[NUM_OF_INPUTS];
static output_t outputs[NUM_OF_OUTPUTS];

static bool debounceInput(input_t* in);

/*
* Functions to initialize the machine
*/
static void init_machine(moore_t* m);
static void init_output(output_id_t id, pin_t pin, write_out write);
static void init_input(input_id_t id, pin_t pin, read_in read, bool debounce);
static void init_state(state_id_t id, next_state_fun_t next, get_output_fun_t out);

/*
* State transition functions related to the states of the coffee machine
*/
static void* standby_next_state(input_t* inputs, void* states, void* par);
static void* supplycup_next_state(input_t* inputs, void* states, void* par);
static void* supplycoffee_next_state(input_t* inputs, void* states, void* par);
static void* ready_next_state(input_t* inputs, void* states, void* par);

/*
* Output functions relates to the states of the coffee machine
*/
static void standby_output(output_t* outputs, void* par);
static void supplycup_output(output_t* outputs, void* par);
static void supplycoffee_output(output_t* outputs, void* par);
static void ready_output(output_t* outputs, void* par);

void initialize(moore_t* m)
{
  int i;
  
  wiringPiSetup();

  //Intializing inputs
  init_input(IN_START_BUTTON, IN_START_BUTTON_PIN, read_input, 1);
  init_input(IN_CUP_READY, IN_CUP_READY_PIN, read_input, 0);

  //Initializing outputs
  init_output(OUT_CUP, OUT_CUP_PIN, write_output);
  init_output(OUT_COFFEE, OUT_COFFEE_PIN, write_output);
  init_output(OUT_READY_LED, OUT_READY_LED_PIN, write_output);
  
  //Registering states
  init_state(STANDBY, standby_next_state, standby_output);
  init_state(SUPPLY_CUP, supplycup_next_state, supplycup_output);
  init_state(SUPPLY_COFFEE, supplycoffee_next_state, supplycoffee_output);
  init_state(READY, ready_next_state, ready_output);
  
  init_machine(m);
}

void init_machine(moore_t* m)
{

	if (m)
	{
		m->num_inputs = NUM_OF_INPUTS;
		m->num_outputs = NUM_OF_OUTPUTS;
		m->num_states = NUM_OF_STATES;
		m->states = states;
		m->inputs = inputs;
		m->outputs = outputs;
		m->current_state = &states[STANDBY];
		m->cycle_limit = TIMELIMIT;
	}
}

void init_state(state_id_t id, next_state_fun_t next, get_output_fun_t out)
{
	states[id].id = id;
	assert(next != NULL && out != NULL);
	states[id].next_state = next;
	states[id].update_output = out;
	printf("State %d initialized\n",states[id].id);
}

void init_input(input_id_t id, pin_t pin, read_in read,bool debounce)
{
  inputs[id].value = 0;
  inputs[id].pin = pin;
  inputs[id].id = id;
  inputs[id].read = read;
  inputs[id].debounce_delay = 20;
  inputs[id].active_debounce = debounce;
  pinMode(pin, INPUT);
  pullUpDnControl(pin, PUD_UP);
  printf("Input %d initialized\n",inputs[id].id);
}

void init_output(output_id_t id, pin_t pin, write_out write)
{
  outputs[id].value = 0;
  outputs[id].pin = pin;
  outputs[id].id = id;
  outputs[id].write = write;
  pinMode(pin, OUTPUT);    
  digitalWrite(pin,0);
  printf("Output %d initialized\n",outputs[id].id);
}

/*STATE FUNCTIONS*/
void* standby_next_state(input_t* inputs, void* m_states, void* par)
{
	assert(inputs != NULL && m_states !=NULL);
	state_t* stvec = (state_t*)m_states;
	bool val_button;
	val_button = inputs[IN_START_BUTTON].value;
	if(val_button)
		return &stvec[SUPPLY_CUP];
	return &stvec[STANDBY];
}

void* supplycup_next_state(input_t* inputs, void* m_states, void* par)
{
	assert(inputs != NULL && m_states !=NULL);
	state_t* stvec = (state_t*)m_states;
	bool val_cup;
	val_cup = inputs[IN_CUP_READY].value;
	if(val_cup)
		return &stvec[SUPPLY_COFFEE];
	return &stvec[SUPPLY_CUP];
}

void* supplycoffee_next_state(input_t* inputs, void* m_states, void* par)
{
	assert(inputs != NULL && m_states !=NULL);
	state_t* stvec = (state_t*)m_states;
	usleep(TIMEOUT);
	return &stvec[READY];
}

void* ready_next_state(input_t* inputs, void* m_states, void* par)
{
	assert(inputs != NULL && m_states !=NULL);
	state_t* stvec = (state_t*)m_states;
	bool val_cup;
	val_cup = inputs[IN_CUP_READY].value;
	if(!val_cup)
		return &stvec[STANDBY];
	return &stvec[READY];
}

void standby_output(output_t* outputs, void* par)
{
  system("clear");
  printf("State: Standby\n");
	outputs[OUT_CUP].value = 0;
	outputs[OUT_COFFEE].value = 0;
	outputs[OUT_READY_LED].value = 0;
}

void supplycup_output(output_t* outputs, void* par)
{
  system("clear");
  printf("State: Suppling Cup\n");
	outputs[OUT_CUP].value = 1;
	outputs[OUT_COFFEE].value = 0;
	outputs[OUT_READY_LED].value = 0;
}

void supplycoffee_output(output_t* outputs, void* par)
{
  system("clear");
  printf("State: Suppling Coffee\n");
	outputs[OUT_CUP].value = 0;
	outputs[OUT_COFFEE].value = 1;
	outputs[OUT_READY_LED].value = 0;
}

void ready_output(output_t* outputs, void* par)
{
  system("clear");
  printf("\nState: Ready\n");
	outputs[OUT_CUP].value = 0;
	outputs[OUT_COFFEE].value = 0;
	outputs[OUT_READY_LED].value = 1;
}

void read_input(void* inp)
{
if(inp)
{
  input_t* in = (input_t*)inp;
  if(in->active_debounce)
  	in->value = !debounceInput(in);
  else
  	in->value = !digitalRead(in->pin);
  printf("Input %d - %d\n", in->id, in->value);
}
}

void write_output(void* outp)
{
	if(outp)
	{
		output_t* out = (output_t*)outp;
      		digitalWrite(out->pin,out->value);
		printf("Output %d - %d\n", out->id, out->value);
	}
}

void begin_cycle(void* p)
{
	printf("Waiting input");
	usleep(500);
}

bool debounceInput(input_t* in)
{
  bool state;
  bool previousState;
  int counter;
  previousState = digitalRead(in->pin);
  for(counter=0; counter < in->debounce_delay; counter++) {
    delay(1);
    state = digitalRead(in->pin);
    if( state != previousState) {
      counter = 0;
      previousState = state; 
    }
  }
  return state;
}
